{$A+,B-,D+,F-,G+,I+,K+,L+,N+,P+,Q-,R-,S+,T-,V-,W-,X+,Y+}
unit MMButton;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Buttons;

CONST LBeveling = 1;
type
  TMMButton = class(TGraphicControl)
  private
    PAutosize: Boolean;
    PBitmap: TBitmap;
    PBitmapUp: TBitmap;
    PBitmapDown: TBitmap;
    PMask: TBitmap;
    PCursorSave: Boolean;
    PCursorAnt: TCursor;
    PSavedHint: Boolean;
    PShowAnt: Boolean;
    PShowHint: Boolean;
    PColorHighlight : TColor;
    PColorDown : TColor;
    procedure AdjustBounds;
    procedure AdjustSize(var W, H: Integer);
    function BevelColor(const AState: TButtonState; const TopLeft: Boolean): TColor;
    procedure BitmapChanged(Sender: TObject);
    procedure Create3DBitmap(Source: TBitmap; const AState: TButtonState; Target: TBitmap);
    procedure SetAutoSize(Value: Boolean);
    procedure SetHighlight(Value: TColor);
    procedure SetDown(Value: TColor);
    procedure SetBitmap(Value: TBitmap);
    procedure SetBitmapDown(Value: TBitmap);
    procedure SetBitmapUp(Value: TBitmap);
    procedure CMDialogChar(var Message: TCMDialogChar); message CM_DIALOGCHAR;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure CMTextChanged(var Message: TMessage); message CM_TEXTCHANGED;
    procedure CMSysColorChange(var Message: TMessage); message CM_SYSCOLORCHANGE;
  protected
    FState: TButtonState;
    procedure DefineProperties(Filer: TFiler); override;
    procedure DrawButtonText(Canvas: TCanvas; const Caption: String; TextBounds: TRect; State: TButtonState); virtual;
    function GetPalette: HPALETTE; override;
    procedure Loaded; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;  X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;  X, Y: Integer); override;
    procedure Paint; override;
    procedure ReadBitmapDownData(Stream: TStream); virtual;
    procedure ReadBitmapUpData(Stream: TStream); virtual;
    procedure WriteBitmapDownData(Stream: TStream); virtual;
    procedure WriteBitmapUpData(Stream: TStream); virtual;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Click; override;
    procedure Invalidate; override;
    function PtInMask(const X, Y: Integer): Boolean; virtual;
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
    property BitmapUp: TBitmap read PBitmapUp;
    property BitmapDown: TBitmap read PBitmapDown;
  published
    property AutoSize: Boolean read PAutosize write SetAutoSize default True;
    property Bitmap: TBitmap read PBitmap write SetBitmap;
    property ColorHighlight : TColor read PColorHighlight write SetHighlight;
    property ColorDown : TColor read PColorDown write SetDown;
    property Caption;
    property Enabled;
    property Font;
    property ParentFont;
    property ShowHint;
    property Visible;
    property OnClick;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
  end;

procedure Register;

implementation

type
  Apair = Array[0..1] of Integer;

function MakeMask(ColorBmp: TBitmap; TransparentColor: TColor): TBitmap;
var R: TRect;
    OldBkColor: TColorRef;
begin
  Result := TBitmap.Create;
  try
    Result.Monochrome := True;
    Result.Width := ColorBmp.Width;
    Result.Height := ColorBmp.Height;
    OldBkColor := SetBkColor(ColorBmp.Canvas.Handle, ColorToRGB(TransparentColor));
    R := Rect(0, 0, ColorBmp.Width, ColorBmp.Height);
    Result.Canvas.CopyMode := cmSrcCopy;
    Result.Canvas.CopyRect(R, ColorBmp.Canvas, R);
    SetBkColor(ColorBmp.Canvas.Handle, OldBkColor);
  except
    Result.Free;
    Raise;
  end;

end;

function DoBordering(Source, NewSource: TBitmap; const OffsetPts: Array of Apair;
  TransparentColor: TColor): TBitmap;
var I, W, H: Integer;
    R, NewR: TRect;
    SmallMask, BigMask, NewSourceMask: TBitmap;
begin
  Result := TBitmap.Create;
  try
    W := Source.Width;
    H := Source.Height;
    R := Rect(0, 0, W, H);

    Result.Monochrome := True;
    Result.Width := W;
    Result.Height := H;

    SmallMask := MakeMask(Source, TransparentColor);
    NewSourceMask := MakeMask(NewSource, TransparentColor);
    BigMask := MakeMask(NewSourceMask, TransparentColor);

    try

      BigMask.Canvas.CopyMode := cmSrcCopy;
      BigMask.Canvas.CopyRect(R, NewSourceMask.Canvas, R);

      for I := Low(OffsetPts) to High(OffsetPts) do
      begin
        if (OffsetPts[I, 0] = 0) and (OffsetPts[I, 1] = 0) then
          Break;
        NewR := R;
        OffsetRect(NewR, OffsetPts[I, 0], OffsetPts[I, 1]);
        BigMask.Canvas.CopyMode := cmSrcAnd;
        BigMask.Canvas.CopyRect(NewR, SmallMask.Canvas, R);
      end;
      BigMask.Canvas.CopyMode := cmSrcCopy;

      with Result do
      begin
        Canvas.CopyMode := cmSrcCopy;
        Canvas.CopyRect(R, NewSourceMask.Canvas, R);
        Canvas.CopyMode := $00DD0228;
        Canvas.CopyRect(R, BigMask.Canvas, R);
        Canvas.CopyMode := cmSrcCopy;
      end;

    finally
      SmallMask.Free;
      NewSourceMask.Free;
      BigMask.Free;
    end;

  except
    Result.Free;
    Raise;
  end;

end;

{ TMMButton }
constructor TMMButton.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  SetBounds(0, 0, 80, 80);
  ControlStyle := [csCaptureMouse, csOpaque];
  PAutosize := True;
  PBitmap := TBitmap.Create;
  PBitmap.OnChange := BitmapChanged;
  PBitmapUp := TBitmap.Create;
  PBitmapDown := TBitmap.Create;
  PMask := nil;
  ParentFont := True;
  FState := bsUp;
  PShowHint := True;
  PColorHighlight := clBtnHighlight;
  PColorDown := clBtnShadow;
{  Caption := ClassName;}
end;

destructor TMMButton.Destroy;
begin
  PBitmap.Free;
  PBitmapUp.Free;
  PBitmapDown.Free;
  PMask.Free;
  inherited Destroy;
end;

procedure TMMButton.Paint;
var W, H: Integer;
    Composite, Mask, Overlay, CurrentBmp: TBitmap;
    R, NewR: TRect;
begin
  if csDesigning in ComponentState then
    with Canvas do
    begin
      Pen.Style := psDash;
      Brush.Style := bsClear;
      Rectangle(0, 0, Width, Height);
    end;

  if (csDesigning in ComponentState) or
    (FState in [bsDisabled, bsExclusive]) then
    FState := bsUp;

  if (FState = bsUp) then CurrentBmp := PBitmapUp
  else CurrentBmp := PBitmapDown;

  if not CurrentBmp.Empty then
  begin

    W := Width;
    H := Height;
    R := ClientRect;
    NewR := R;

    Composite := TBitmap.Create;
    Overlay := TBitmap.Create;

    try
      with Composite do
      begin
        Width := W;
        Height := H;
        Canvas.CopyMode := cmSrcCopy;
        Canvas.CopyRect(R, Self.Canvas, R);
      end;

      with Overlay do
      begin
        Width := W;
        Height := H;
        Canvas.CopyMode := cmSrcCopy;
        Canvas.Brush.Color := PBitmap.TransparentColor;
        Canvas.FillRect(R);
        if FState = bsDown then
          OffsetRect(NewR, 1, 1);
        Canvas.CopyRect(NewR, CurrentBmp.Canvas, R);
      end;

      Mask := MakeMask(Overlay, PBitmap.TransparentColor);
      try

        Composite.Canvas.CopyMode := cmSrcAnd;
        Composite.Canvas.CopyRect(R, Mask.Canvas, R);


        Overlay.Canvas.CopyMode := $00220326;
        Overlay.Canvas.CopyRect(R, Mask.Canvas, R);


        Composite.Canvas.CopyMode := cmSrcPaint;
        Composite.Canvas.CopyRect(R, Overlay.Canvas, R);

        Canvas.CopyMode := cmSrcCopy;
        Canvas.CopyRect(R, Composite.Canvas, R);

      finally
        Mask.Free;
      end;

    finally
      Composite.Free;
      Overlay.Free;
    end;

  end;

  if Length(Caption) > 0 then
  begin
    Canvas.Font := Self.Font;
    R := CLIENTRECT;
    DrawButtonText(Canvas, Caption, R, FState);
  end;

end;

function TMMButton.PtInMask(const X, Y: Integer): Boolean;
begin
  Result := True;
  if PMask <> nil then
    Result := (PMask.Canvas.Pixels[X, Y] = clBlack);
end;

procedure TMMButton.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
var Clicked: Boolean;
begin
  inherited MouseDown(Button, Shift, X, Y);
  if (Button = mbLeft) and Enabled then
  begin
      Clicked := PtInMask(X, Y);

    if Clicked then
    begin
      FState := bsDown;
      Repaint;
    end;
  end;
end;

procedure TMMButton.MouseMove(Shift: TShiftState; X, Y: Integer);
var NewState: TButtonState;
    InMask: Boolean;
begin
  inherited MouseMove(Shift, X, Y);
  InMask := PtInMask(X, Y);

  if PShowHint and not InMask then
  begin

    if not PSavedHint then
    begin

      ParentShowHint := False;
      PShowAnt := ShowHint;
      ShowHint := False;
      PSavedHint := True;
    end;
  end
  else IF not InMask then
  begin
    if not PCursorSave then
    begin
      PCursorAnt := Cursor;
      Cursor := crDefault;
      PCursorSave := True;
    end;
  end
  else
  begin
    if PSavedHint then
    begin

      ShowHint := PShowAnt;
      PSavedHint := False;
    end;
    if PCursorSave then
    begin
      Cursor := PCursorAnt;
      PCursorSave := False;
    end;
  end;

end;

procedure TMMButton.MouseUp(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
var
  DoClick: Boolean;
begin
  inherited MouseUp(Button, Shift, X, Y);
  DoClick := PtInMask(X, Y);
  if (FState = bsDown) then
    begin
      FState := bsUp;
      Repaint;
    end;
    if DoClick then Click;
end;

procedure TMMButton.Click;
begin
  inherited Click;
end;

function TMMButton.GetPalette: HPALETTE;
begin
  Result := PBitmap.Palette;
end;

procedure TMMButton.SetBitmap(Value: TBitmap);
begin
  PBitmap.Assign(Value);
end;

procedure TMMButton.SetBitmapUp(Value: TBitmap);
begin
  PBitmapUp.Assign(Value);
end;

procedure TMMButton.SetBitmapDown(Value: TBitmap);
begin
  PBitmapDown.Assign(Value);
end;

procedure TMMButton.BitmapChanged(Sender: TObject);
var OldCursor: TCursor;
    W, H: Integer;
begin
  AdjustBounds;

  if not ((csReading in ComponentState) or (csLoading in ComponentState)) then
  begin
    if PBitmap.Empty then
    begin

      SetBitmapUp(nil);
      SetBitmapDown(nil);
    end
    else
    begin
      W := PBitmap.Width;
      H := PBitmap.Height;
      OldCursor := Screen.Cursor;
      Screen.Cursor := crHourGlass;
      try
        if (PBitmapUp.Width <> W) or (PBitmapUp.Height <> H) or
          (PBitmapDown.Width <> W) or (PBitmapDown.Height <> H) then
        begin
          PBitmapUp.Width := W;
          PBitmapUp.Height := H;
          PBitmapDown.Width := W;
          PBitmapDown.Height := H;
        end;
        Create3DBitmap(PBitmap, bsUp, PBitmapUp);
        Create3DBitmap(PBitmap, bsDown, PBitmapDown);

        PMask.Free;
        PMask := MakeMask(PBitmapUp, PBitmap.TransparentColor);

      finally
        Screen.Cursor := OldCursor;
      end;
    end;
  end;
  Invalidate;

end;

procedure TMMButton.CMDialogChar(var Message: TCMDialogChar);
begin
  with Message do
    if IsAccel(CharCode, Caption) and Enabled then
    begin
      Click;
      Result := 1;
    end else
      inherited;
end;

procedure TMMButton.CMFontChanged(var Message: TMessage);
begin
  Invalidate;
end;

procedure TMMButton.CMTextChanged(var Message: TMessage);
begin
  Invalidate;
end;

procedure TMMButton.CMSysColorChange(var Message: TMessage);
begin
  BitmapChanged(Self);
end;

function TMMButton.BevelColor(const AState: TButtonState; const TopLeft: Boolean): TColor;
begin
  if (AState = bsUp) then
  begin
    if TopLeft then Result := PColorHighlight
    else Result := PColorDown;
  end
  else { bsDown }
  begin
    if TopLeft then Result := PColorDown
    else Result := PColorHighlight;
  end;
end;


procedure TMMButton.Create3DBitmap(Source: TBitmap; const AState: TButtonState; Target: TBitmap);
type OutlineOffsetPts = Array[1..3, 0..1, 0..12] of Apair;
const
  OutlinePts: OutlineOffsetPts =
    ( (((1,-1),(1,0),(1,1),(0,1),(-1,1),(0,0),(0,0),(0,0),(0,0),(0,0),(0,0),(0,0),(0,0)),
       ((-1,0),(-1,-1),(0,-1),(0,0),(0,0),(0,0),(0,0),(0,0),(0,0),(0,0),(0,0),(0,0),(0,0))),
      (((2,-2),(2,-1),(2, 0),(2, 1),(2, 2),(1, 2),(0, 2),(-1,2),(-2,2),(0,0),(0,0),(0,0),(0,0)),
       ((-2,1),(-2,0),(-2,-1),(-2,-2),(-1,-2),(0,-2),(1,-2),(0,0),(0,0),(0,0),(0,0),(0,0),(0,0))),
      (((3,-3),(3,-2),(3,-1),(3,0),(3,1),(3,2),(3,3),(2,3),(1,3),(0,3),(-1,3),(-2,3),(-3,3)),
       ((-3,2),(-3,1),(-3,0),(-3,-1),(-3,-2),(-3,-3),(-2,-3),(-1,-3),(0,-3),(1,-3),(2,-3),(0,0),(0,0)))
    );
var I, J, W, H, Outlines: Integer;
    R: TRect;
    OutlineMask, Overlay, NewSource: TBitmap;
begin
  if (Source = nil) or (Target = nil) then
    Exit;

  W := Source.Width;
  H := Source.Height;
  R := Rect(0, 0, W, H);

  Overlay := TBitmap.Create;
  NewSource := TBitmap.Create;

  try

    NewSource.Width := W;
    NewSource.Height := H;


    Target.Canvas.CopyMode := cmSrcCopy;
    Target.Canvas.CopyRect(R, Source.Canvas, R);

    Overlay.Width := W;
    Overlay.Height := H;

    Outlines := LBeveling;
      Inc(Outlines);

    for I := 1 to Outlines do
    begin
      with NewSource.Canvas do
      begin
        CopyMode := cmSrcCopy;
        CopyRect(R, Target.Canvas, R);
      end;

      for J := 0 to 1 do
      begin
        if (AState = bsDown) and (I = Outlines) and (J = 0) then
          Continue;

        OutlineMask := DoBordering(Source, NewSource, OutlinePts[I, J],
                        PBitmap.TransparentColor);
        try
          with Overlay.Canvas do
          begin
            if (I = Outlines)  then
              Brush.Color := clBlack
            else
              Brush.Color := BevelColor(AState, (J = 1));
            CopyMode := $0030032A; { PSna }
            CopyRect(R, OutlineMask.Canvas, R);
          end;

          with Target.Canvas do
          begin

            CopyMode := cmSrcAnd; { DSa }
            CopyRect(R, OutlineMask.Canvas, R);

            CopyMode := cmSrcPaint; { DSo }
            CopyRect(R, Overlay.Canvas, R);
            CopyMode := cmSrcCopy;
          end;

        finally
          OutlineMask.Free;
        end;

      end;
    end;

  finally
    Overlay.Free;
    NewSource.Free;

  end;
end;

procedure TMMButton.DrawButtonText(Canvas: TCanvas; const Caption: String;
  TextBounds: TRect; State: TButtonState);
var
  CString: array[0..255] of Char;
begin
  StrPCopy(CString, Caption);
  Canvas.Brush.Style := bsClear;
  if State = bsDown then OffsetRect(TextBounds, 1, 1);
  DrawText(Canvas.Handle, CString, -1, TextBounds,
      DT_CENTER or DT_VCENTER or DT_SINGLELINE);
end;

procedure TMMButton.Loaded;
var BigMask: TBitmap;
    R: TRect;
begin
  inherited Loaded;
  if (PBitmap <> nil) and (PBitmap.Width > 0) and (PBitmap.Height > 0) then
  begin
    PMask.Free;
    PMask := MakeMask(PBitmap, PBitmap.TransparentColor);
    BigMask := MakeMask(PBitmapUp, PBitmap.TransparentColor);
    try
      R := Rect(0, 0, PBitmap.Width, PBitmap.Height);
      PMask.Canvas.CopyMode := cmSrcAnd;
      PMask.Canvas.CopyRect(R, BigMask.Canvas, R);
    finally
      BigMask.Free;
    end;
  end;
end;


procedure TMMButton.DefineProperties(Filer: TFiler);
begin
  inherited DefineProperties(Filer);

  Filer.DefineBinaryProperty('BitmapUp', ReadBitmapUpData, WriteBitmapUpData, not PBitmapUp.Empty);

  Filer.DefineBinaryProperty('BitmapDown', ReadBitmapDownData, WriteBitmapDownData, not PBitmapDown.Empty)
end;

procedure TMMButton.ReadBitmapUpData(Stream: TStream);
begin
  PBitmapUp.LoadFromStream(Stream);
end;

procedure TMMButton.WriteBitmapUpData(Stream: TStream);
begin
  PBitmapUp.SaveToStream(Stream);
end;

procedure TMMButton.ReadBitmapDownData(Stream: TStream);
begin
  PBitmapDown.LoadFromStream(Stream);
end;

procedure TMMButton.WriteBitmapDownData(Stream: TStream);
begin
  PBitmapDown.SaveToStream(Stream);
end;

procedure TMMButton.AdjustBounds;
begin
  SetBounds(Left, Top, Width, Height);
end;

procedure TMMButton.AdjustSize(var W, H: Integer);
begin
  if not (csReading in ComponentState) and PAutosize and not PBitmap.Empty then
  begin
    W := PBitmap.Width;
    H := PBitmap.Height;
  end;
end;

procedure TMMButton.SetAutoSize(Value: Boolean);
begin
  if PAutosize <> Value then
  begin
    PAutosize := Value;
    AdjustBounds;
  end;
end;

procedure TMMButton.SetHighlight(Value: TColor);
begin
  if PColorHighlight <> Value then
  begin
    PColorHighlight := Value;
    Repaint;
  end;
end;

procedure TMMButton.SetDown(Value: TColor);
begin
  if PColorDown <> Value then
  begin
    PColorDown := Value;
    Repaint;
  end;
end;

procedure TMMButton.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
var W, H: Integer;
begin
  W := AWidth;
  H := AHeight;
  AdjustSize(W, H);
  inherited SetBounds(ALeft, ATop, W, H);
end;

procedure TMMButton.Invalidate;
var R: TRect;
begin
  if (Visible or (csDesigning in ComponentState)) and
    (Parent <> nil) and Parent.HandleAllocated then
  begin
    R := BoundsRect;
    InvalidateRect(Parent.Handle, @R, True);
  end;
end;

procedure Register;
begin
  RegisterComponents('WMB', [TMMButton]);
end;

end.
